unit HCLists;

interface

uses
  Classes;

type
{ TSortedList object }
  TSortedListMethod = procedure (Item: Pointer) of object;
  PSortedList = ^TSortedList;
  TSortedList = class( TList )
  private
    FDuplicates: boolean;
  public
    constructor Create( DuplicateKey: boolean );
    function    Add(Item: Pointer): Integer; virtual;
    function    IndexOf(Item: Pointer): Integer; virtual;
    function    Search(Key: Pointer; var Index: Integer): Boolean; virtual;
    function    Find(Item: Pointer): Pointer; virtual;
    procedure   ForEach(Action: TSortedListMethod); virtual;
    function    KeyOf(Item: Pointer): Pointer; virtual;
{!} procedure   PassMeTheKey(Item: Pointer; Here: Pointer); virtual; {!! PShortString !!}
{!} function    Compare(Key1, Key2: Pointer): Integer; virtual; {!! PShortString !!}
  end;

  PSortedDWORDKeyList = ^TSortedDWORDKeyList;
  TSortedDWORDKeyList = class( TSortedList )
  public
    procedure   PassMeTheKey(Item: Pointer; Here: Pointer); override;
    function    Compare(Key1, Key2: Pointer): Integer; override;
  end;

implementation

uses
  Windows,Consts;

{ TSortedList object }
constructor TSortedList.Create( DuplicateKey: boolean );
begin
  inherited Create;
  FDuplicates:=DuplicateKey;
end;

function TSortedList.Add(Item: Pointer): Integer;
var
  I: Integer;
  Key: ShortString;
begin
  PassMeTheKey(Item,@Key);
  If NOT Search(@Key, I) OR FDuplicates Then Begin
    Insert(I, Item);
  End Else I:= -1;
  Result:= I;
end;

function TSortedList.IndexOf(Item: Pointer): Integer;
var
  I: integer;
  Key: ShortString;
begin
  Result:= -1;
  PassMeTheKey(Item,@Key);
  If Search(@Key, I) Then Begin
    If FDuplicates Then Begin
      While (I < Count) AND (Item <> Items[I]) DO Inc(I);
    End;
    If I < Count Then Result := I;
  End;
end;

function TSortedList.Search(Key: Pointer; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
  KeyI: ShortString;
begin
  Search := False;
  L := 0;
  H := Count - 1;
  While L <= H Do Begin
    I := (L + H) shr 1;
    PassMeTheKey(Items[I],@KeyI);
    C := Compare(@KeyI, Key);
    If C < 0 Then
      L := I + 1
    Else Begin
      H := I - 1;
      If C = 0 Then Begin
        Search := True;
        If NOT FDuplicates then L := I;
      End;
    End;
  End;
  Index := L;
end;

function TSortedList.Find(Item: Pointer): Pointer;
Var
  KeyIndex: integer;
  RecP: Pointer;
  KeyP: ShortString;
  Key: ShortString;
begin
  RecP:=NIL;
  PassMeTheKey(Item,@Key);
  If Search(@Key,KeyIndex) Then Begin
    If (KeyIndex>0) AND (KeyIndex<Count) Then Begin
      RecP:=Items[KeyIndex];
      PassMeTheKey(RecP,@KeyP);
      If Compare(@Key,@KeyP)<>0 Then RecP:=NIL;
    End;
  End;
  Result:=RecP;
end;

procedure TSortedList.ForEach(Action: TSortedListMethod);
var
  I: integer;
begin
  For I:=0 To Count-1 Do Action(Items[I]);
end;

function TSortedList.KeyOf(Item: Pointer): Pointer;
begin
  Result:= Item;
end;

function TSortedList.Compare(Key1, Key2: Pointer): Integer;
begin
  If (Key1=NIL) OR (Key2=NIL) Then Begin
    If (Key1=NIL) AND (Key2=NIL) Then
      Result:=0
    Else Begin
      If Key1=NIL Then
        Result:=-1
      Else Result:=1;
    End;
  End Else Begin
    If ShortString(Key1^)=ShortString(Key2^) Then
      Result:=0
    Else Begin
      If ShortString(Key1^)<ShortString(Key2^) Then
        Result:=-1
      Else Result:=1;
    End;
  End;
end;

procedure TSortedList.PassMeTheKey(Item: Pointer; Here: Pointer);
begin
  If Here=NIL Then
    Error(sNoAddress,-1)
  Else ShortString(Here^):=ShortString(KeyOf(Item)^);
end;

function TSortedDWORDKeyList.Compare(Key1, Key2: Pointer): Integer;
begin
  If (Key1=NIL) OR (Key2=NIL) Then Begin
    If (Key1=NIL) AND (Key2=NIL) Then
      Result:=0
    Else Begin
      If Key1=NIL Then
        Result:=-1
      Else Result:=1;
    End;
  End Else Begin
    If DWORD(Key1^)=DWORD(Key2^) Then
      Result:=0
    Else Begin
      If DWORD(Key1^)<DWORD(Key2^) Then
        Result:=-1
      Else Result:=1;
    End;
  End;
end;

procedure TSortedDWORDKeyList.PassMeTheKey(Item: Pointer; Here: Pointer);
begin
  If Here=NIL Then
    Error(sNoAddress,-1)
  Else DWORD(Here^):=DWORD(KeyOf(Item)^);
end;

end.
